# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import subprocess
import os


class HVMetric():
    """Main Halstead Volume class."""

    input = ''

    def __init__(self, input):
        """Initialize class."""
        if len(input) == 0:
            raise ValueError('Empty file for analysis')
        self.input = input

    def value(self):
        """Run Halstead Volume analaysis"""
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..',
                               'halsteadvolume/target/java-project-1.0-SNAPSHOT.jar'))
        if not os.path.isfile(path):
            hvpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'halsteadvolume'))
            subprocess.run(['mvn', 'clean', 'package'], cwd=hvpath)
        result = subprocess.run(['java', '-jar', path, self.input], stdout=subprocess.PIPE)
        out = result.stdout.decode('utf-8')
        res = result = {'data': [{'file': self.input, 'halsteadvolume': float(out)}]}
        return res



import javalang


class InstanceOf:

    def __init__(self):
        pass

    def __traverse_node(self, node):
        """
        Travers over AST tree finds instance_of and .isInstance().
        :param filename:
        :return:
        List of code lines
        """
        lines = []
        for path, node_elem in node.filter(javalang.tree.BinaryOperation):
            if node_elem.operator == 'instanceof':
                code_line = node_elem.operandl.position.line or node_elem.operandr.position.line
                lines.append(code_line)
        for path, node_elem in node.filter(javalang.tree.MethodInvocation):
            if node_elem.member == 'isInstance':
                lines.append(node_elem.position.line)

        return lines

    def __file_to_ast(self, filename: str) -> javalang.ast.Node:
        """
        Takes path to java class file and returns AST Tree
        :param filename:
        :return: Tree
        """
        with open(filename, encoding='utf-8') as file:
            res = javalang.parse.parse(file.read())
        return res

    # flake8: noqa: C901
    def value(self, filename: str):
        tree = self.__file_to_ast(filename)
        return self.__traverse_node(tree)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import javalang
from typing import List
from enum import Enum


class NodeType(Enum):
    VAR = 1      # Variable Declaration
    METHOD = 2   # Method Declaration


class VarMiddle:
    '''
    Returns lines in the file where variables declared in the middle
    of the method
    '''

    def __init__(self):
        pass

    def __file_to_ast(self, filename: str) -> javalang.ast.Node:
        '''Takes path to java class file and returns AST Tree'''
        with open(filename, encoding='utf-8') as file:
            tree = javalang.parse.parse(file.read())

        return tree

    def __check_var_declaration(self, pos: int, line_to_node_map, empty_lines: List[int]) -> bool:
        '''
        Check that VAR declaration line is near the method declaration
        Args:
            pos (int): Line number we are going to check.
            line_to_node_map (Map Int Node): Map line number to node object

        Returns:
            bool: True if VAR declaration is near method declaration
        '''

        if line_to_node_map.get(pos) != NodeType.VAR:
            raise ValueError('Variable declaration line is expected!')
        i = pos - 1
        while i > 0:
            if line_to_node_map.get(i) == NodeType.METHOD:
                return True
            if i in empty_lines:
                i -= 1
                continue
            if line_to_node_map.get(i) is None:
                return False
            i -= 1
        raise ValueError('Method declaration is not found')

    def __get_empty_lines(self, tree: javalang.tree.CompilationUnit) -> List[int]:
        '''Figure out lines that are either empty or multiline statements'''
        lines_with_nodes = [
            node.position.line for path, node in tree
            if hasattr(node, 'position') and node.position is not None
        ]
        max_line = max(lines_with_nodes)
        return set(range(1, max_line + 1)).difference(lines_with_nodes)

    def value(self, filename: str) -> List[int]:
        ''''''
        line_to_node_map = {}
        tree = self.__file_to_ast(filename)
        empty_lines = self.__get_empty_lines(tree)

        m_decls = tree.filter(javalang.tree.MethodDeclaration)
        v_decls = tree.filter(javalang.tree.LocalVariableDeclaration)
        m_poss = [(e.position.line, NodeType.METHOD) for _, e in m_decls]
        v_poss = [(e.position.line, NodeType.VAR) for _, e in v_decls]
        line_to_node_map = dict(m_poss + v_poss)

        return [
            line for line, _ in v_poss
            if not self.__check_var_declaration(line, line_to_node_map, empty_lines)
        ]



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import javalang
from typing import List, Callable, Optional, Any


class BlockType:
    FOR = javalang.tree.ForStatement      # FOR Block Statement
    IF = javalang.tree.IfStatement        # IF Block Statement


class NestedBlocks:
    '''
    Returns lines in the file where
    nested FOR/IF blocks are located
    '''

    def __init__(self, max_depth: int, block_type=BlockType.FOR):
        self.max_depth = max_depth
        self.block_type = block_type

    def __file_to_ast(self, filename: str) -> javalang.ast.Node:
        '''Takes path to java class file and returns AST Tree'''
        with open(filename, encoding='utf-8') as file:
            tree = javalang.parse.parse(file.read())

        return tree

    def __for_node_depth(
        self,
        tree: javalang.ast.Node,
        max_depth: int,
        for_links: List = [],
        for_before: int = 0
    ) -> None:
        '''
        Takes AST tree and returns list of "FOR" AST nodes of depth greater
        or equal than max_depth
        '''
        if (type(tree) == self.block_type):
            for_before += 1
            if for_before >= max_depth:
                for_links += [tree]

        for child in tree.children:
            nodes_arr = child if isinstance(child, list) else [child]
            for node in nodes_arr:
                if not hasattr(node, 'children'):
                    continue
                self.__for_node_depth(node, max_depth, for_links, for_before)

    def __fold_traverse_tree(
        self,
        root: javalang.ast.Node,
        f: Callable[[javalang.ast.Node], Optional[Any]]
    ) -> [Any]:
        '''
        Traverse AST tree and apply function to each node
        Accumulate results in the list and return
        '''
        res = []
        v = f(root)
        if v is not None:
            res.append(v)
        for child in root.children:
            nodes_arr = child if isinstance(child, list) else [child]
            for node in nodes_arr:
                if not hasattr(node, 'children'):
                    continue
                res += self.__fold_traverse_tree(node, f)
        return res

    def value(self, filename: str) -> List[int]:
        '''Return line numbers in the file where patterns are found'''
        tree = self.__file_to_ast(filename)
        for_links = []
        self.__for_node_depth(
            tree,
            max_depth=self.max_depth,
            for_links=for_links
        )

        def find_line_position(node: javalang.ast.Node) -> int:
            if hasattr(node, '_position'):
                return node._position.line
            else:
                None
        n_lines = [
            self.__fold_traverse_tree(for_node, find_line_position)
            for for_node in for_links
        ]
        return list(map(min, n_lines))



#!/usr/bin/env python
# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""The main entry point. Invoke as `aibolit' or `python -m aibolit'.
"""
import argparse
import os
import queue as queue
import sys
from pathlib import Path

from aibolit.patterns.nested_blocks.nested_blocks import NestedBlocks, BlockType
from aibolit.patterns.string_concat.string_concat import StringConcatFinder
from aibolit.patterns.var_middle.var_middle import VarMiddle
from aibolit import __version__


def main():
    exit_status = -1
    try:
        depth_for = 2
        depth_if = 2
        parser = argparse.ArgumentParser(
            description='Find the pattern which has the largest impact on readability'
        )
        parser.add_argument(
            '--filename',
            help='path for Java file')
        parser.add_argument('--version', action='version',
                            version='%(prog)s {version}'.format(version=__version__))

        args = parser.parse_args(args=None if sys.argv[1:] else ['--help'])

        if args:
            java_file = str(Path(os.getcwd(), args.filename))

            order_queue = queue.Queue()
            order_queue.put([
                VarMiddle().value(java_file),
                'variable declaration in the middle of the function'])
            order_queue.put([
                NestedBlocks(depth_for, block_type=BlockType.FOR).value(java_file),
                'nested for cycle with depth = {}'.format(depth_for)])
            order_queue.put([
                NestedBlocks(depth_if, block_type=BlockType.IF).value(java_file),
                'nested if condition with depth = 2'.format(depth_if)])
            order_queue.put([
                StringConcatFinder().value(java_file),
                'string concatenation with operator +'])

            lines, output_string = order_queue.get()
            if not lines:
                print('Your code is perfect in aibolit\'s opinion')
            for line in lines:
                if line:
                    print('Line {}. Low readability due to: {}'.format(
                        line,
                        output_string
                    ))
            order_queue.task_done()
            exit_status = 0
    except KeyboardInterrupt:
        exit_status = -1
    sys.exit(exit_status)


if __name__ == '__main__':
    main()



import javalang
import re


class StringConcatFinder:

    def __init__(self):
        pass

    def remove_comments(self, string):
        # remove all occurrences streamed comments (/*COMMENT */) from string
        string = re.sub(re.compile(r"/\*.*?\*/", re.DOTALL), "",
                        string)
        # remove all occurrence single-line comments (//COMMENT\n ) from string
        string = re.sub(re.compile(r"//.*?\n"), "",
                        string)
        return string

    def __file_to_ast(self, filename: str) -> javalang.ast.Node:
        """
        Takes path to java class file and returns AST Tree
        :param filename:
        :return: Tree
        """
        with open(filename, encoding='utf-8') as file:
            res = self.remove_comments(file.read())

        return res

    def value(self, filename: str):
        text = self.__file_to_ast(filename)
        pattern_without_quote_first = re.compile(r'(?<=([\w])\+)\"[\w]+\"')
        pattern_with_quote_first = re.compile(r'(?<=([\w]\")\+)[\w]+')
        lines = []
        for iter, line in enumerate(text.splitlines()):
            t_str = line.replace(' ', '')
            if re.search(pattern_without_quote_first, t_str) \
                    or re.search(pattern_with_quote_first, t_str):
                lines.append(iter)
        return lines



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from setuptools import setup, find_packages
import aibolit

setup(
    name='aibolit',
    version=aibolit.__version__,
    description=aibolit.__doc__.strip(),
    long_description='Defect Detection with Machine Learning in Mind',
    url='https://github.com/yegor256/aibolit',
    download_url='https://github.com/yegor256/aibolit',
    author=aibolit.__author__,
    author_email='yegor256@gmail.com',
    license=aibolit.__licence__,
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'aibolit = aibolit.__main__:main'
        ],
    },
    extras_require={},
    install_requires=[],
    tests_require=[],
    classifiers=[
        'Programming Language :: Python',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development',
        'Topic :: Utilities'
    ],
)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.instanceof.instance_of import InstanceOf
from pathlib import Path


class TestInstanceOf(TestCase):
    cur_file_dir = Path(os.path.realpath(__file__)).parent
    testClass = InstanceOf()

    def test_empty(self):
        file = str(Path(self.cur_file_dir, 'Empty.java'))
        assert len(self.testClass.value(file)) == 0

    def test_instance_of(self):
        file = str(Path(self.cur_file_dir, 'InstanceOfSample.java'))
        assert len(self.testClass.value(file)) == 1

    def test_instance(self):
        file = str(Path(self.cur_file_dir, 'InstanceSample.java'))
        assert len(self.testClass.value(file)) == 1

    def test_instance_of_different_methods(self):
        file = str(Path(self.cur_file_dir, 'InstanceOfSampleDifferentMethods.java'))
        assert len(self.testClass.value(file)) == 2

    def test_instance_different_methods(self):
        file = str(Path(self.cur_file_dir, 'InstanceSampleDifferentMethods.java'))
        assert len(self.testClass.value(file)) == 2

    def test_instance_of_several(self):
        file = str(Path(self.cur_file_dir, 'InstanceOfSampleSeveral.java'))
        assert len(self.testClass.value(file)) == 3

    def test_instance_several(self):
        file = str(Path(self.cur_file_dir, 'InstanceSampleSeveral.java'))
        assert len(self.testClass.value(file)) == 2

    def test_instance_in_method_chaining(self):
        file = str(Path(self.cur_file_dir, 'InstanceSampleChain.java'))
        assert len(self.testClass.value(file)) == 2

    def test_both(self):
        file = str(Path(self.cur_file_dir, 'InstanceBoth.java'))
        assert len(self.testClass.value(file)) == 2



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import javalang
from typing import List


class VarDeclarationDistance:
    '''
    Returns lines where variable first time used but declared more than
    specific number of lined before
    '''

    def __init__(self, lines_th: int):
        self.__lines_th = lines_th

    def __file_to_ast(self, filename: str) -> javalang.ast.Node:
        '''Takes path to java class file and returns AST Tree'''
        with open(filename, encoding='utf-8') as file:
            tree = javalang.parse.parse(file.read())

        return tree

    def __process_node(self, node):
        line = node.position.line if hasattr(node, 'position') and node.position is not None else None
        qualifier = node.qualifier if hasattr(node, 'qualifier') else None
        member = node.member if hasattr(node, 'member') else None
        name = node.name if hasattr(node, 'name') else None
        return {
            "line": line,
            "name": qualifier or member or name,
            "ntype": type(node)
        }

    def __tree_to_list(self, tree: javalang.tree.CompilationUnit) -> List[object]:
        '''Convert AST tree to list of object'''
        items = [self.__process_node(node) for path, node in tree if node is not None]

        # fill missed line numbers
        last_line_number = None

        for item in items:
            if (item['line']) is not None:
                last_line_number = item['line']
                continue
            item['line'] = last_line_number

        return items

    def __get_empty_lines(self, tree: javalang.tree.CompilationUnit) -> List[int]:
        '''Figure out lines that are either empty or multiline statements'''
        lines_with_nodes = [
            node.position.line for path, node in tree
            if hasattr(node, 'position') and node.position is not None
        ]
        max_line = max(lines_with_nodes)
        return set(range(1, max_line + 1)).difference(lines_with_nodes)

    def __group_vars_by_method(self, items: List[object]) -> List[object]:
        '''
        Group variables by method scope and calculate for each the declaration
        line and first usage line
        '''
        var_scopes = []
        vars = {}
        for item in items:
            if item['ntype'] in [javalang.tree.MethodDeclaration]:
                vars = {}
                var_scopes += [vars]
                continue

            if (item['ntype'] == javalang.tree.VariableDeclarator):
                vars[item['name']] = {'decl': item['line'], 'first_usage': None}
                continue

            if (item['ntype'] == javalang.tree.VariableDeclarator):
                vars[item['name']] = {'decl': item['line']}
                continue

            if item['name'] in vars.keys() and vars[item['name']]['first_usage'] is None:
                vars[item['name']]['first_usage'] = item['line']

        return var_scopes

    def __line_diff(self, usage_line: int, declaration_line: int, empty_lines: List[int]) -> int:
        '''
        Calculate line difference between variable declaration and first usage
        taking into account empty lines
        '''
        lines_range = set(range(declaration_line + 1, usage_line))
        return len(lines_range.difference(empty_lines))

    def value(self, filename: str) -> List[int]:
        ''''''

        tree = self.__file_to_ast(filename)
        items = self.__tree_to_list(tree)
        var_scopes = self.__group_vars_by_method(items)
        empty_lines = self.__get_empty_lines(tree)

        violations = []
        for scope in var_scopes:
            for var in scope:
                if scope[var]['first_usage'] is None:
                    continue

                line_diff = self.__line_diff(
                    scope[var]['first_usage'],
                    scope[var]['decl'],
                    empty_lines
                )
                if line_diff < self.__lines_th:
                    continue

                violations.append(scope[var]['first_usage'])

        return violations



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from statistics import variance
import re


class SpacesCounter:

    def __init__(self):
        pass

    def remove_comments(self, string):
        # remove all occurrences streamed comments (/*COMMENT */) from string
        string = re.sub(re.compile(r"/\*.*?\*/", re.DOTALL), "",
                        string)
        # remove all occurrence single-line comments (//COMMENT\n ) from string
        string = re.sub(re.compile(r"//.*?\n"), "",
                        string)
        return string

    def __file_to_tokens(self, filename: str):
        """
        Takes path to java class file and returns tokens

        :param filename: file name
        :return: list of counted spaces
        """
        with open(filename, encoding='utf-8') as file:
            text = self.remove_comments(file.read())
            lines = []
            for x in text.splitlines():
                line = x.replace('\n', '').replace('\t', '    ')
                if line:
                    lines.append(line)

        return lines

    def value(self, filename: str):
        lines = self.__file_to_tokens(filename)
        spaces_per_line = []
        prev_left = 0
        prev_right = len(lines[0])
        for i, line in enumerate(lines):
            first_non_space_symbol_pos = len(line) - len(line.lstrip(' '))
            spaces_per_line.append(
                [i,
                 # count the difference between the position of first symbol in cur string and
                 # the position of first symbol of previous string
                 first_non_space_symbol_pos - prev_left,
                 # count the difference between the position of last symbol in cur string and
                 # the position of last symbol of previous string
                 len(line) - prev_right
                 ])
            prev_left = first_non_space_symbol_pos
            prev_right = len(line)

        left_space_variance = variance([x[1] for x in spaces_per_line])
        right_space_variance = variance([x[2] for x in spaces_per_line])
        max_left_space_diff = max([abs(x[1]) for x in spaces_per_line])
        max_right_space_diff = max([abs(x[2]) for x in spaces_per_line])
        return [left_space_variance, right_space_variance,
                max_left_space_diff, max_right_space_diff]
